from py_compile import compile  # pylint: disable=W0622
from os import makedirs, walk, sep
from os.path import join, dirname, exists, isfile, isdir
from shutil import copyfile, copytree
from click import echo, style

from equipment.Command.AbstractCommand import AbstractCommand


class CompileCommand(AbstractCommand):
    def run(self, dist: str) -> None:
        try:
            echo(style(f'Compiling project into {dist}...', fg='green'))

            ignore_dirs = ['__pycache__', 'dist', 'tests', 'equipment']

            for root, dirs, files in walk('.'):
                # Filter out ignored directories
                dirs[:] = [d for d in dirs if d not in ignore_dirs]
                for file in files:
                    if file.endswith('.py'):
                        source_path = join(root, file)
                        dest_path = join('.', dist, root.replace('/', sep).replace('\\', sep), f'{file}c')
                        makedirs(dirname(dest_path), exist_ok=True)
                        compile(source_path, dest_path)

            include_sources = ['config', 'database', 'storage', '.coveragerc', '.editorconfig',
                            '.env', '.env.example', '.gitignore', 'pyproject.toml', 'README.md']

            for source in include_sources:
                source_env_path = join('.', source)
                dest_env_path = join('.', dist, source.replace( '/', sep).replace('\\', sep))
                if exists(source_env_path):
                    if isfile(source_env_path):
                        copyfile(source_env_path, dest_env_path)
                    if isdir(source_env_path):
                        copytree(source_env_path, dest_env_path, dirs_exist_ok=True)


            echo(style(f'Project successfully compiled on {dist}!', fg='green'))
        except Exception as e:
            echo(style(f'Fatal error: {e}', fg='red'))
            return



import os
from logging import Handler, FileHandler, Logger, NullHandler, StreamHandler, getLogger, getLevelNamesMapping
from logging.handlers import TimedRotatingFileHandler
from pythonjsonlogger.jsonlogger import JsonFormatter
from python_sqlite_log_handler import SQLiteLogHandler
from equipment.Log.AbstractLogger import AbstractLogger


class LoggerFactory(AbstractLogger):
    logger: Logger
    base_path: str
    config: dict

    def __init__(self, name: str, base_path: str, config: dict):
        self.base_path = base_path
        self.config = config

        if not hasattr(self, 'logger'):
            logger = getLogger(name)
            _clear_logger(logger)

            logger.setLevel(
                getLevelNamesMapping()[str(
                    self.config['level']).upper().strip()
                ]
            )

            handlers = self._get_handlers_list()

            for handler in handlers:
                logger.addHandler(handler)

            self.logger = logger

    def debug(self, *args, **kwargs) -> None:
        kwargs.setdefault('stacklevel', 2)
        self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs) -> None:
        kwargs.setdefault('stacklevel', 2)
        self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs) -> None:
        kwargs.setdefault('stacklevel', 2)
        self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs) -> None:
        kwargs.setdefault('stacklevel', 2)
        self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs) -> None:
        kwargs.setdefault('stacklevel', 2)
        self.logger.critical(*args, **kwargs)

    def _get_handlers_list(self) -> list[Handler]:
        channel = self.config['channel']

        if channel is None or channel == 'None' or channel == 'null':
            return [NullHandler()]

        handlers = []
        channels = list(
            self.config['channels']['stack']['channels']
        ) if channel == 'stack' else [channel]

        for channel in channels:
            handlers.append(
                self._get_channel_handler(
                    channel,
                    self.config['channels'][channel]
                )
            )

        return handlers

    def _get_channel_handler(self, channel: str, channel_config: dict) -> Handler:
        if channel == 'single':
            handler = FileHandler(os.path.join(
                self.base_path,
                str(channel_config['filename'])
            ))
        elif channel == 'daily':
            handler = TimedRotatingFileHandler(
                filename=os.path.join(
                    self.base_path,
                    str(channel_config['filename'])
                ),
                when=str(channel_config['when']),
                interval=int(channel_config['interval']),
                backupCount=int(channel_config['backupCount'])
            )
        elif channel == 'console':
            handler = StreamHandler(channel_config['stream'])
        elif channel == 'sqlite':
            handler = SQLiteLogHandler(
                db_path=os.path.join(
                    self.base_path,
                    str(channel_config['filename'])
                ),
                table_name=channel_config['table_name'] if 'table_name' in channel_config else 'logs'
            )
        else:
            raise ValueError(f'Invalid channel name: {channel}')

        if 'formatter' in channel_config and str(channel_config['formatter']).strip().lower() == 'json':
            handler.setFormatter(
                JsonFormatter(
                    self.config['formatters']['json']['format'],
                    json_indent=self.config['formatters']['json']['indent']
                )
            )

        return handler


def _clear_logger(logger: Logger) -> None:
    handlers = logger.handlers
    for handler in handlers:
        handler.close()
        logger.removeHandler(handler)



from typing import TYPE_CHECKING
from datetime import datetime
from redis import Redis
from rq import Queue
from equipment.Queue.AbstractQueue import AbstractQueue

if TYPE_CHECKING:
    from equipment.Log.AbstractLogger import AbstractLogger


class RedisQueue(AbstractQueue):
    _queue: Queue

    def __init__(self, config: dict, log: 'AbstractLogger'):
        self.log = log
        self.config = config

        if not hasattr(self, '_queue'):
            redis = Redis(
                host=config['host'],
                port=int(config['port']),
                db=int(config['db']),
                # TODO: add more configurable properties
            )

            self._queue = Queue(connection=redis)

    def push(self, method, *args, **kwargs) -> bool:
        try:
            self._queue.enqueue(
                f=method,
                args=args,
                kwargs=kwargs
            )

            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def push_at(self, date: 'datetime', method, *args, **kwargs) -> bool:
        try:
            self._queue.enqueue_at(
                datetime=date,
                f=method,
                args=args,
                kwargs=kwargs
            )

            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False



from typing import TYPE_CHECKING
from equipment.Storage.AbstractStorage import AbstractStorage
from equipment.Storage.LocalStorage import LocalStorage
from equipment.Storage.S3Storage import S3Storage

if TYPE_CHECKING:
    from equipment.Log.AbstractLogger import AbstractLogger


class StorageFactory(AbstractStorage):
    storage: AbstractStorage

    def __init__(self, config: dict, base_path: str, log: 'AbstractLogger'):
        if not hasattr(self, 'storage'):
            if config['disk'] == 'local':
                self.storage = LocalStorage(
                    config['disks']['local'],
                    base_path,
                    log
                )
            elif config['disk'] == 's3':
                self.storage = S3Storage(config['disks']['s3'], log)
            else:
                raise ValueError(f'Unknown storage type: {config["disk"]}')

    def path(self, file: str) -> str:
        return self.storage.path(file)

    def write(self, file: str, data: str) -> bool:
        return self.storage.write(file, data)

    def read(self, file: str) -> str:
        return self.storage.read(file)

    def exists(self, file: str) -> bool:
        return self.storage.exists(file)

    def remove(self, file: str) -> bool:
        return self.storage.remove(file)

    def move(self, source: str, destination: str) -> bool:
        return self.storage.move(source, destination)

    def list(self, path: str) -> list:
        return self.storage.list(path)



from typing import TYPE_CHECKING
import boto3
from botocore.exceptions import ClientError

from equipment.Storage.AbstractStorage import AbstractStorage

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
    from equipment.Log.AbstractLogger import AbstractLogger


class S3Storage(AbstractStorage):
    s3: 'S3Client'

    def __init__(self, config: dict, log: 'AbstractLogger'):
        self.config = config
        self.log = log

    def client(self) -> 'S3Client':
        if not hasattr(self, 's3'):
            self.s3 = boto3.client(
                's3',
                endpoint_url=self.config['endpoint'],
                aws_access_key_id=self.config['access_key'],
                aws_secret_access_key=self.config['secret_key'],
                region_name=self.config['region']
            )

        return self.s3

    def _get_full_path(self, file: str) -> str:
        prefix = self.config['prefix'] if 'prefix' in self.config else ''
        prefix = '' if prefix is None or prefix == 'None' or prefix == 'null' else prefix
        if prefix and not prefix.endswith('/'):
            prefix += '/'
        return f"{prefix}{file}"

    def path(self, file: str) -> str:
        return self._get_full_path(file)

    def write(self, file: str, data: str) -> bool:
        try:
            s3_path = self._get_full_path(file)

            self.client().put_object(
                Bucket=self.config['bucket'],
                Key=s3_path,
                Body=data
            )
            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def read(self, file: str) -> str:
        if not self.exists(file):
            raise FileNotFoundError(file)

        try:
            s3_path = self._get_full_path(file)

            response = self.client().get_object(
                Bucket=self.config['bucket'],
                Key=s3_path
            )
            return response['Body'].read().decode('utf-8')
        except Exception as e:
            self.log.error(e, exc_info=True)
            raise

    def exists(self, file: str) -> bool:
        try:
            s3_path = self._get_full_path(file)
            self.client().head_object(
                Bucket=self.config['bucket'],
                Key=s3_path
            )
            return True
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                return False
            self.log.error(e, exc_info=True)
            return False
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def remove(self, file: str) -> bool:
        try:
            s3_path = self._get_full_path(file)

            self.client().delete_object(
                Bucket=self.config['bucket'],
                Key=s3_path
            )
            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def move(self, source: str, destination: str) -> bool:
        try:
            source_path = self._get_full_path(source)
            dest_path = self._get_full_path(destination)

            # Copy the object to the new location
            self.client().copy_object(
                Bucket=self.config['bucket'],
                CopySource={
                    'Bucket': self.config['bucket'], 'Key': source_path},
                Key=dest_path
            )

            # Delete the original
            self.client().delete_object(
                Bucket=self.config['bucket'],
                Key=source_path
            )

            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def list(self, path: str) -> list:
        try:
            s3_path = self._get_full_path(path)
            if not s3_path.endswith('/'):
                s3_path += '/'

            response = self.client().list_objects_v2(
                Bucket=self.config['bucket'],
                Prefix='' if s3_path == '/' else s3_path,
            )

            # Extract just filenames from the full paths
            contents = response.get('Contents', [])
            files = []

            for obj in contents:
                if 'Key' not in obj:
                    continue

                # Skip directories (objects that end with /)
                key = obj['Key']
                if not key.endswith('/'):
                    # Remove the prefix and path to get just the filename
                    name = key.replace(s3_path, '', 1)
                    if name:  # Only add non-empty names
                        files.append(name)

            return files
        except Exception as e:
            self.log.error(e, exc_info=True)
            raise



import os
from typing import TYPE_CHECKING
from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker

if TYPE_CHECKING:
    from sqlalchemy import Engine, TextClause
    from sqlalchemy.orm.session import Session
    from equipment.Log.AbstractLogger import AbstractLogger


def _create_sqlite_url(config: dict, base_path: str) -> str:
    schema = config['schema']
    database = '' if config['database'] == ':memory:' else os.path.join(
        base_path,
        config['database']
    )

    return f"{schema}://" if database == '' else f"{schema}:///{database}"


def _create_mysql_url(config: dict) -> str:
    return f"{config['schema']}://{config['username']}:{config['password']}@{config['host']}:{config['port']}/{config['database']}?charset={config['charset']}"


def _create_postgresql_url(config: dict) -> str:
    return f"{config['schema']}://{config['username']}:{config['password']}@{config['host']}:{config['port']}/{config['database']}"


class SQLAlchemyFactory:
    url: str
    session_maker: sessionmaker
    engine: 'Engine'

    def __init__(self, config: dict, base_path: str, log: 'AbstractLogger'):
        if not hasattr(self, 'session_maker'):
            if config['connection'] == 'sqlite':
                self.url = _create_sqlite_url(config['connections']['sqlite'], base_path)
            elif config['connection'] == 'mysql':
                self.url = _create_mysql_url(config['connections']['mysql'])
            elif config['connection'] == 'postgresql':
                self.url = _create_postgresql_url(config['connections']['postgresql'])
            else:
                raise ValueError(f'Unknown queue connection type: {config["connection"]}')

            self.engine = create_engine(self.url)
            self.session_maker = sessionmaker(bind=self.engine)

    def session(self) -> 'Session':
        return self.session_maker()

    @staticmethod
    def text(query: str) -> 'TextClause':
        return text(query)



import unittest
from dependency_injector.containers import DynamicContainer
from equipment import Equipment, equipment
from equipment.Log.LoggerFactory import LoggerFactory
from equipment.Queue.QueueFactory import QueueFactory
from equipment.Storage.StorageFactory import StorageFactory
from equipment.Database.SQLAlchemyFactory import SQLAlchemyFactory
from runtype import isa

TEST_CONFIG = {
    "app": {
        "name": "test",
        "env": "testing",
    },
    "log": {
        "level": "DEBUG",
        "channel": None,
        "formatter": None,
    },
    "storage": {
        "disk": "local",
        "disks": {
            "local": {
                "path": "storage"
            }
        }
    },
    "queue": {
        "connection": "sync",
        "connections": {
            "sync"
        }
    },
    "database": {
        "connection": "sqlite",
        "connections": {
            "sqlite": {
                "schema": "sqlite",
                "database": ":memory:",
            }
        }
    }
}


class EquipmentTest(unittest.TestCase):
    def test_equipment_method(self):
        instance = equipment()
        self.assertTrue(isa(instance, DynamicContainer))

    def test_default_singletons(self):
        instance = Equipment()

        instance.config.from_dict(TEST_CONFIG)

        self.assertTrue(isa(instance.log(), LoggerFactory))
        self.assertTrue(isa(instance.storage(), StorageFactory))
        self.assertTrue(isa(instance.queue(), QueueFactory))
        self.assertTrue(isa(instance.database(), SQLAlchemyFactory))


if __name__ == '__main__':
    unittest.main()



from equipment.Log.AbstractLogger import AbstractLogger


class NullLogger(AbstractLogger):
    def debug(self, *args, **kwargs) -> None:
        pass

    def info(self, *args, **kwargs) -> None:
        pass

    def warning(self, *args, **kwargs) -> None:
        pass

    def error(self, *args, **kwargs) -> None:
        pass

    def critical(self, *args, **kwargs) -> None:
        pass



from typing import TYPE_CHECKING
from equipment.Queue.AbstractQueue import AbstractQueue
from equipment.Queue.SyncQueue import SyncQueue
from equipment.Queue.RedisQueue import RedisQueue

if TYPE_CHECKING:
    from datetime import datetime
    from equipment.Log.AbstractLogger import AbstractLogger


class QueueFactory(AbstractQueue):
    queue: AbstractQueue

    def __init__(self, config: dict, log: 'AbstractLogger'):
        if not hasattr(self, 'queue'):
            if config['connection'] == 'sync':
                self.queue = SyncQueue(log)
                return

            if config['connection'] == 'redis':
                self.queue = RedisQueue(config['connections']['redis'], log)
                return

            raise ValueError(f'Unknown queue connection type: {config["connection"]}')

    def push(self, method, *args, **kwargs) -> bool:
        return self.queue.push(method, *args, **kwargs)

    def push_at(self, date: 'datetime', method, *args, **kwargs) -> bool:
        return self.queue.push_at(date, method, *args, **kwargs)



from os import sep
from os.path import abspath, dirname, isdir, join
from codecs import open  # pylint: disable=W0622
from shutil import copyfile, copytree, ignore_patterns, rmtree
from zipfile import ZipFile
from io import BytesIO
from click import confirm, echo, style
from requests import get
from equipment.Command.AbstractCommand import AbstractCommand

class NewProjectCommand(AbstractCommand):
    def run(self, name: str, path: str) -> None:
        project_template_path = f'{abspath(dirname(__file__))}{sep}_project'

        if isdir(project_template_path):
            rmtree(project_template_path)

        try:
            response = get('https://github.com/rogervila/equipment/archive/refs/heads/main.zip', timeout=60)
            with ZipFile(BytesIO(response.content)) as zip_file:
                zip_file.extractall(project_template_path)
        except Exception as e:
            echo(style(f'Could not download project template: {e}', fg='red'))
            return

        project_template_path = f'{project_template_path}{sep}equipment-main{sep}project'

        if not isdir(project_template_path):
            echo(style(f'Project template not found in "{project_template_path}". If the error persists, please open an issue', fg='red'))
            return

        try:
            echo(style(f'Creating "{name}" project on {path}...', fg='green'))

            project_path = f'{path}{sep}{name}'
            already_exists = isdir(project_path)
            confirmation = confirm(
                f'Directory "{project_path}" already exists. Do you want to override it?' # autopep8: off
            ) if already_exists else True

            if not confirmation:
                echo(style('Skip', fg='yellow'))
                return None

            copytree(
                project_template_path,
                project_path,
                symlinks=True,
                dirs_exist_ok=True,
                ignore=ignore_patterns(
                    '*.py[co]',
                    '__pycache__',
                    '.git',
                    'htmlcov',
                    'equipment',
                    'venv',
                    '.env',
                    '.coverage',
                    '*.log',
                    '*.lock',
                    '*.sqlite*',
                    '.DS_Store',
                )
            )

            copyfile(
                join(project_template_path, '.env.example'),
                join(project_path, '.env')
            )

            pyproject_path = join(project_path, 'pyproject.toml')

            with open(pyproject_path, 'r') as file:
                pyproject_content = file.read()

            pyproject_content = pyproject_content.replace('PROJECT_NAME', name)

            with open(pyproject_path, 'w') as file:
                file.write(pyproject_content)

            echo(style(f'Project "{name}" successfully created on {project_path}!', fg='green'))
        except Exception as e:
            echo(style(f'Fatal error: {e}', fg='red'))
            return



import os
import click
from .NewProjectCommand import NewProjectCommand
from .CompileCommand import CompileCommand


@click.group()
@click.version_option()
def main() -> None:
    pass  # @click decorators handle the commands defined on this file


@main.command()
@click.argument('name')
def new(name: str) -> None:
    NewProjectCommand().run(
        name=name,
        path=os.getcwd(),
    )


@main.command()
@click.argument('dist')
def compile(dist: str) -> None:  # pylint: disable=W0622
    CompileCommand().run(
        dist=dist,
    )



import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import datetime


class AbstractQueue(abc.ABC):
    def push(self, method, *args, **kwargs) -> bool:
        raise NotImplementedError

    def push_at(self, date: 'datetime', method, *args, **kwargs) -> bool:
        raise NotImplementedError



import os
from glob import glob

from dependency_injector.containers import DeclarativeContainer
from dependency_injector.providers import ThreadSafeSingleton as Singleton, Configuration
from dotenv import load_dotenv
from cachetools import cached

from equipment.Database.SQLAlchemyFactory import SQLAlchemyFactory
from equipment.Log.LoggerFactory import LoggerFactory
from equipment.Queue.QueueFactory import QueueFactory
from equipment.Storage.StorageFactory import StorageFactory


def _load_environment(base_path: str) -> None:
    env_file = os.path.join(base_path, '.env')

    if os.path.isfile(env_file):
        load_dotenv(env_file)


def _config_files_list(base_path: str, extension: str) -> list[str]:
    return glob(
        os.path.join(
            base_path,
            'config',
            extension
        )
    )


class Equipment(DeclarativeContainer):
    # Ensure that no unexpected arguments are passed to the __new__ constructor.
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    @classmethod
    @cached({})
    def make(cls, base_path: tuple[str, None] = None):
        base_path = base_path if base_path is not None else os.getcwd()
        _load_environment(base_path)
        instance = cls()

        for file in _config_files_list(base_path, '*.ini'):
            instance.config.from_ini(file)

        for file in _config_files_list(base_path, '*.yaml'):
            instance.config.from_yaml(file)

        for file in _config_files_list(base_path, '*.json'):
            instance.config.from_json(file)

        instance.config.base_path.from_value(base_path)

        assert instance.config.base_path() == base_path

        return instance

    config = Configuration()

    log = Singleton(
        LoggerFactory, config.app.name, config.base_path, config.log
    )

    queue = Singleton(
        QueueFactory, config.queue, log
    )

    storage = Singleton(
        StorageFactory, config.storage, config.base_path, log
    )

    database = Singleton(
        SQLAlchemyFactory, config.database, config.base_path, log
    )


def equipment(base_path: tuple[str, None] = None) -> Equipment:
    return Equipment.make(base_path)


def cli() -> None:
    from equipment.Command import main as _cli
    _cli()



from typing import TYPE_CHECKING
from equipment.Queue.AbstractQueue import AbstractQueue

if TYPE_CHECKING:
    from datetime import datetime
    from equipment.Log.AbstractLogger import AbstractLogger


class SyncQueue(AbstractQueue):
    def __init__(self, log: 'AbstractLogger'):
        self.log = log

    def push(self, method, *args, **kwargs) -> bool:
        try:
            method(*args, **kwargs)
            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def push_at(self, date: 'datetime', method, *args, **kwargs) -> bool:
        # SyncQueue ignores specific dates
        return self.push(method, *args, **kwargs)



import abc


class AbstractLogger(abc.ABC):
    def debug(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def info(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def warning(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def error(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def critical(self, *args, **kwargs) -> None:
        raise NotImplementedError



import codecs
import os
from typing import TYPE_CHECKING

from equipment.Storage.AbstractStorage import AbstractStorage

if TYPE_CHECKING:
    from equipment.Log.AbstractLogger import AbstractLogger


class LocalStorage(AbstractStorage):
    def __init__(self, config: dict, base_path: str, log: 'AbstractLogger'):
        self.config = config
        self.base_path = base_path
        self.log = log

    def _local_storage_path(self) -> str:
        return os.path.join(
            self.base_path,
            self.config['path']
        )

    def path(self, file: str) -> str:
        return os.path.join(
            self._local_storage_path(),
            file
        )

    def write(self, file: str, data: str) -> bool:
        try:
            file = self.path(file)

            os.makedirs(
                os.path.dirname(file),
                exist_ok=True
            )

            with codecs.open(file, 'w') as filepath:
                filepath.write(data)

            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def read(self, file: str) -> str:
        if not self.exists(file):
            raise FileNotFoundError(file)

        file = self.path(file)

        self.log.debug('file: ' + file)

        with codecs.open(file, 'r') as filepath:
            return filepath.read()

    def exists(self, file: str) -> bool:
        return os.path.isfile(self.path(file))

    def remove(self, file: str) -> bool:
        try:
            os.remove(self.path(file))
            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def move(self, source: str, destination: str) -> bool:
        try:
            os.rename(
                self.path(source),
                self.path(destination)
            )
            return True
        except Exception as e:
            self.log.error(e, exc_info=True)
            return False

    def list(self, path: str) -> list:
        path = os.path.join(self._local_storage_path(), path)

        if not os.path.isdir(path):
            raise NotADirectoryError(f'{path} is not a directory')

        return [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file))]
